import mock
import unittest

import redis
from pyramid import testing

from kinto.core.permission.testing import PermissionTest
from kinto_redis import permission as redis_backend


class RedisPermissionTest(PermissionTest, unittest.TestCase):
    backend = redis_backend
    settings = {
        'permission_url': '',
        'permission_pool_size': 10
    }

    def setUp(self):
        super(RedisPermissionTest, self).setUp()
        self.client_error_patcher = [
            mock.patch.object(
                self.permission._client,
                'execute_command',
                side_effect=redis.RedisError),
            mock.patch.object(
                self.permission._client,
                'pipeline',
                side_effect=redis.RedisError)]

    def test_config_is_taken_in_account(self):
        config = testing.setUp(settings=self.settings)
        config.add_settings({'permission_url': 'redis://:pass@db.loc:1234/5'})
        backend = self.backend.load_from_config(config)
        self.assertDictEqual(
            backend.settings,
            {'host': 'db.loc', 'password': 'pass', 'db': 5, 'port': 1234})

    def test_timeout_is_passed_to_redis_client(self):
        config = testing.setUp(settings=self.settings)
        config.add_settings({'permission_pool_timeout': '1.5'})
        backend = self.backend.load_from_config(config)
        self.assertEqual(backend._client.connection_pool.timeout, 1.5)



# -*- coding: utf-8 -*-
import mock
import redis
import unittest
from pyramid import testing

from kinto.core.storage import exceptions
from kinto.core.storage.testing import StorageTest

from kinto_redis import storage as redisbackend


class RedisStorageTest(StorageTest, unittest.TestCase):
    backend = redisbackend
    settings = {
        'storage_pool_size': 50,
        'storage_url': ''
    }

    def setUp(self):
        super(RedisStorageTest, self).setUp()
        self.client_error_patcher = mock.patch.object(
            self.storage._client.connection_pool, 'get_connection',
            side_effect=redis.RedisError('connection error'))

    def test_config_is_taken_in_account(self):
        config = testing.setUp(settings=self.settings)
        config.add_settings({'storage_url': 'redis://:blah@store.loc:7777/6'})
        backend = self.backend.load_from_config(config)
        self.assertDictEqual(
            backend.settings,
            {'host': 'store.loc', 'password': 'blah', 'db': 6, 'port': 7777})

    def test_timeout_is_passed_to_redis_client(self):
        config = testing.setUp(settings=self.settings)
        config.add_settings({'storage_pool_timeout': '1.5'})
        backend = self.backend.load_from_config(config)
        self.assertEqual(backend._client.connection_pool.timeout, 1.5)

    def test_backend_error_is_raised_anywhere(self):
        with mock.patch.object(self.storage._client, 'pipeline',
                               side_effect=redis.RedisError):
            StorageTest.test_backend_error_is_raised_anywhere(self)

    def test_get_all_handle_expired_values(self):
        record = '{"id": "foo"}'.encode('utf-8')
        mocked_smember = mock.patch.object(self.storage._client, "smembers",
                                           return_value=['a', 'b'])
        mocked_mget = mock.patch.object(self.storage._client, "mget",
                                        return_value=[record, None])
        with mocked_smember:
            with mocked_mget:
                self.storage.get_all(**self.storage_kw)  # not raising

    def test_errors_logs_stack_trace(self):
        self.client_error_patcher.start()

        with mock.patch('kinto.core.storage.logger.exception') as exc_handler:
            with self.assertRaises(exceptions.BackendError):
                self.storage.get_all(**self.storage_kw)

        self.assertTrue(exc_handler.called)



from __future__ import absolute_import

import re

from kinto.core.permission import PermissionBase
from kinto_redis.storage import create_from_config, wrap_redis_error


class Permission(PermissionBase):
    """Permission backend implementation using Redis.

    Enable in configuration::

        kinto.permission_backend = kinto_redis.permission

    *(Optional)* Instance location URI can be customized::

        kinto.permission_url = redis://localhost:6379/2

    A threaded connection pool is enabled by default::

        kinto.permission_pool_size = 50

    :noindex:
    """

    def __init__(self, client, *args, **kwargs):
        super(Permission, self).__init__(*args, **kwargs)
        self._client = client

    @property
    def settings(self):
        return dict(self._client.connection_pool.connection_kwargs)

    def initialize_schema(self, dry_run=False):
        # Nothing to do.
        pass

    def _decode_set(self, results):
        return set([r.decode('utf-8') for r in results])

    @wrap_redis_error
    def flush(self):
        self._client.flushdb()

    @wrap_redis_error
    def add_user_principal(self, user_id, principal):
        user_key = 'user:%s' % user_id
        self._client.sadd(user_key, principal)

    @wrap_redis_error
    def remove_user_principal(self, user_id, principal):
        user_key = 'user:%s' % user_id
        self._client.srem(user_key, principal)
        if self._client.scard(user_key) == 0:
            self._client.delete(user_key)

    def remove_principal(self, principal):
        with self._client.pipeline() as pipe:
            user_keys = self._client.scan_iter(match='user:*')
            for user_key in user_keys:
                pipe.srem(user_key, principal)
            pipe.execute()

    @wrap_redis_error
    def get_user_principals(self, user_id):
        # Fetch the groups the user is in.
        user_key = 'user:%s' % user_id
        members = self._decode_set(self._client.smembers(user_key))
        # Fetch the groups system.Authenticated is in.
        group_authenticated = self._decode_set(self._client.smembers('user:system.Authenticated'))
        return members | group_authenticated

    @wrap_redis_error
    def add_principal_to_ace(self, object_id, permission, principal):
        permission_key = 'permission:%s:%s' % (object_id, permission)
        self._client.sadd(permission_key, principal)

    @wrap_redis_error
    def remove_principal_from_ace(self, object_id, permission, principal):
        permission_key = 'permission:%s:%s' % (object_id, permission)
        self._client.srem(permission_key, principal)
        if self._client.scard(permission_key) == 0:
            self._client.delete(permission_key)

    @wrap_redis_error
    def get_object_permission_principals(self, object_id, permission):
        permission_key = 'permission:%s:%s' % (object_id, permission)
        members = self._client.smembers(permission_key)
        return self._decode_set(members)

    @wrap_redis_error
    def get_accessible_objects(self, principals, bound_permissions=None, with_children=True):
        principals = set(principals)

        if bound_permissions:
            keys = ['permission:%s:%s' % op for op in bound_permissions]
        else:
            keys = ['permission:*']
        perms_by_id = dict()
        for key_pattern in keys:
            matched = handle_with_children(self._client.scan_iter(match=key_pattern),
                                           bound_permissions, with_children)
            for key in matched:
                authorized = self._decode_set(self._client.smembers(key))
                if len(authorized & principals) > 0:
                    _, obj_id, permission = key.decode('utf-8').split(':', 2)
                    perms_by_id.setdefault(obj_id, set()).add(permission)

        return perms_by_id

    @wrap_redis_error
    def get_authorized_principals(self, bound_permissions):
        keys = ['permission:%s:%s' % (o, p) for (o, p) in bound_permissions]
        if keys:
            return self._decode_set(self._client.sunion(*list(keys)))
        return set()

    @wrap_redis_error
    def get_objects_permissions(self, objects_ids, permissions=None):
        objects_perms = []
        for object_id in objects_ids:
            if permissions:
                keys = ['permission:%s:%s' % (object_id, permission)
                        for permission in permissions]
            else:
                keys = [key.decode('utf-8') for key in self._client.scan_iter(
                    match='permission:%s:*' % object_id)]

            with self._client.pipeline() as pipe:
                for permission_key in keys:
                    pipe.smembers(permission_key)

                results = pipe.execute()

            permissions = {}
            for i, result in enumerate(results):
                permission = keys[i].split(':', 2)[-1]
                principals = self._decode_set(result)
                if principals:
                    permissions[permission] = principals
            objects_perms.append(permissions)
        return objects_perms

    @wrap_redis_error
    def replace_object_permissions(self, object_id, permissions):
        keys = ['permission:%s:%s' % (object_id, permission)
                for permission in permissions]
        with self._client.pipeline() as pipe:
            for key in keys:
                pipe.delete(key)
                permission = key.split(':', 2)[-1]
                principals = permissions[permission]
                if len(principals) > 0:
                    pipe.sadd(key, *principals)
            pipe.execute()

    @wrap_redis_error
    def delete_object_permissions(self, *object_id_list):
        with self._client.pipeline() as pipe:
            for object_id in object_id_list:
                keys = list(self._client.scan_iter(
                    match='permission:%s:*' % object_id))
                if len(keys) > 0:
                    pipe.delete(*keys)
            pipe.execute()


def handle_with_children(keys, bound_permissions, with_children):
    results = set()
    if bound_permissions and not with_children:
        regexps = [
            re.compile(('permission:%s:%s' % (object_id, permission)).replace('*', '[^/]+'))
            for object_id, permission in bound_permissions]

        for key in keys:
            decoded_key = key.decode('utf-8')
            matches = [True for r in regexps if r.match(decoded_key)]
            if matches:
                results.add(key)
        return list(results)
    return keys


def load_from_config(config):
    client = create_from_config(config, prefix='permission_')
    return Permission(client)



# -*- coding: utf-8 -*-
import json
import uuid
import unittest
from datetime import datetime
from contextlib import contextmanager

import mock
from pyramid import testing

from kinto.core import initialization
from kinto.core.events import ResourceChanged, ACTIONS

from kinto_redis.storage import create_from_config


@contextmanager
def broken_redis():
    from redis import StrictRedis
    old = StrictRedis.lpush

    def push(*args, **kwargs):
        raise Exception('boom')

    StrictRedis.lpush = push
    yield
    StrictRedis.lpush = old


UID = str(uuid.uuid4())


class Resource(object):
    record_id = UID
    timestamp = 123456789


class ViewSet(object):
    def get_name(*args, **kw):
        return 'collection'


class Service(object):
    viewset = ViewSet()


class Match(object):
    cornice_services = {'watev': Service()}
    pattern = 'watev'


class Request(object):
    path = '/1/bucket/collection/'
    prefixed_userid = 'tarek'
    matchdict = {'id': UID}
    registry = matched_route = Match()
    current_resource_name = 'bucket'


class ListenerCalledTest(unittest.TestCase):

    def setUp(self):
        self.config = testing.setUp()
        self.config.add_settings({'events_pool_size': 1,
                                  'events_url': 'redis://localhost:6379/0'})
        self._redis = create_from_config(self.config, prefix='events_')
        self._size = 0

        self.sample_event = ResourceChanged({'action': ACTIONS.CREATE.value},
                                            [],
                                            Request())

    def _save_redis(self):
        self._size = self._redis.llen('kinto.core.events')

    def has_redis_changed(self):
        return self._redis.llen('kinto.core.events') > self._size

    def notify(self, event):
        self._save_redis()
        self.config.registry.notify(event)

    @contextmanager
    def redis_listening(self):
        config = self.config
        listener = 'kinto_redis.listeners'

        # setting up the redis listener
        with mock.patch.dict(config.registry.settings,
                             [('event_listeners', listener),
                              ('event_listeners.redis.pool_size', '1')]):
            initialization.setup_listeners(config)
            config.commit()
            yield

    def test_redis_is_notified(self):
        with self.redis_listening():
            # let's trigger an event
            self.notify(self.sample_event)
            self.assertTrue(self.has_redis_changed())

        # okay, we should have the first event in Redis
        last = self._redis.lpop('kinto.core.events')
        last = json.loads(last.decode('utf8'))
        self.assertEqual(last['action'], ACTIONS.CREATE.value)

    def test_notification_is_broken(self):
        with self.redis_listening():
            # an event with a bad JSON should silently break and send nothing
            # date time objects cannot be dumped
            event2 = ResourceChanged({'action': ACTIONS.CREATE.value,
                                      'somedate': datetime.now()},
                                     [],
                                     Request())
            self.notify(event2)
            self.assertFalse(self.has_redis_changed())

    def test_redis_is_broken(self):
        with self.redis_listening():
            # if the redis call fails, same deal: we should ignore it
            self._save_redis()

            with broken_redis():
                self.config.registry.notify(self.sample_event)

            self.assertFalse(self.has_redis_changed())



import codecs
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    README = f.read()

with codecs.open(os.path.join(here, 'CHANGELOG.rst'), encoding='utf-8') as f:
    CHANGELOG = f.read()

REQUIREMENTS = [
    'kinto>=4.0.0',
    'redis',
]

setup(name='kinto-redis',
      version='1.1.0',
      description='Kinto Redis',
      long_description=README + '\r\n' + CHANGELOG,
      license='Apache License (2.0)',
      classifiers=[
          "Programming Language :: Python",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
          "License :: OSI Approved :: Apache Software License"
      ],
      keywords="web services",
      author='Mozilla Services',
      author_email='services-dev@mozilla.com',
      url='https://github.com/Kinto/kinto-redis',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=REQUIREMENTS)
